"""
Sparse Grid generator.

Author: Panos Tsilifis
Date: 8/25/2017

"""

__all__ = ['QuadratureRule']


import numpy as np
import math
import itertools as itl
from scipy import misc
from scipy.linalg import eigh_tridiagonal
from . import MonicPoly


class QuadratureRule(object):


	_rule_list = ['GL', 'GH', 'NC', 'CC', 'GLag', 'custom_pdf']
	_rule = None


	@property 
	def rule(self):
		return self._rule

	@rule.setter
	def rule(self, value):
		assert value in self._rule_list, 'Input rule is not (yet) supported. Currently available options are : \n' + str(self._rule_list)
		self._rule = value

	def __init__(self, rule = 'GH'):
		assert rule in self._rule_list, 'Input rule is not (yet) supported. Currently available options are : \n' + str(self._rule_list)
		self._rule = rule

	def mi_terms(self, order, dim):
		""" matrix of basis terms
		Input
		:order: PCE order
		:dim: PCE dimension        
		"""
		q_num = [int(misc.comb(dim+i-1, i)) for i in range(order+1)]
		mul_ind = np.array(np.zeros(dim, dtype = int), dtype = int)
		mul_ind = np.vstack([mul_ind, np.eye(dim, dtype = int)])
		I = np.eye(dim, dtype = int)
		ind = [1] * dim
		terms = []
		for j in range(1,order):
			ind_new = []
			for i in range(dim):
				a0 = np.copy(I[int(np.sum(ind[:i])):,:])
				a0[:,i] += 1
				mul_ind = np.vstack([mul_ind, a0])
				ind_new += [a0.shape[0]]
			ind = ind_new
			I = np.copy(mul_ind[np.sum(q_num[:j+1]):,:])
			terms = [list(mul_ind[i,:]) for i in range(mul_ind.shape[0]) if mul_ind[i,:].min() > 0]
		return np.array(terms)


	def GaussHermite(self, n, odd = False):
		assert n > 0
		assert isinstance(n, int)
		if odd:
			n = 2 ** n - 1
		d = np.sqrt(np.arange(n))[1:]
		#H = np.diag(d, -1) + np.diag(d, 1)
		#[x, v] = np.linalg.eigh(H)
		[x, v] = eigh_tridiagonal(np.zeros(n), d)
		if odd:
			x[(n-1)/2] = 0.
		w = v[0,:] ** 2
		rule = {'x': x, 'w': w}
		return rule


	def GaussLegendre(self, n, odd = False):
		assert n > 0
		assert isinstance(n, int)
		if odd:
			n = 2 ** n - 1
		d = np.sqrt([i**2/((2.*i+1.)*(2*i-1.)) for i in range(1,n)])
		#H = np.diag(d, -1) + np.diag(d, 1)
		#[x, v] = np.linalg.eigh(H)
		[x, v] = eigh_tridiagonal(np.zeros(n), d)
		w = v[0,:] ** 2
		if (x.shape[0]-1) % 2 == 0:
			x[(x.shape[0]-1) / 2] = 0.
		rule = {'x': x, 'w': w}
		return rule


	def GaussLaguerre(self, n, odd = False):
		assert n > 0
		assert isinstance(n, int)
		if odd:
			n = 2 ** n - 1
		d = [i for i in range(1,n)]
		alpha = [2*i+1 for i in range(n)]
		#H = np.diag(alpha) + np.diag(d, -1) + np.diag(d, 1)
		#[x, v] = np.linalg.eigh(H)
		[x, v] = eigh_tridiagonal(alpha, d)
		w = v[0,:] ** 2
		rule = {'x': x, 'w': w}
		return rule


	def Gauss(self, n, pdf, supp, odd = False):
		assert n > 0
		assert isinstance(n, int)
		if odd:
			n = 2 ** n - 1
		monic = MonicPoly(n, pdf, supp)
		alpha, beta = monic.recurr_coeffs()
		#H = np.diag(alpha) + np.diag(np.sqrt(beta), -1) + np.diag(np.sqrt(beta), 1)
		#[x, v] = np.linalg.eigh(H)
		[x, v] = eigh_tridiagonal(alpha, np.sqrt(beta))
		w = v[0,:] ** 2
		rule = {'x': x, 'w': w}
		return rule


	def Trapezoidal(self, l):
		assert l > 0
		assert isinstance(l, int)
		if l == 1:
			x = np.array([0.])
			w = np.array([1.])
		else:
			n = 2 ** (l - 1) + 1
			x = np.array([i*2**(2-l)-1. for i in range(n)])
			w = np.array([2 ** (1.-l) for i in range(n)])
			w[0] = w[0] / 2.
			w[-1] = w[-1] / 2.
		rule = {'x': x, 'w': w}
		return rule


	def ClenshawCurtis(self, l):
		assert l > 0
		assert isinstance(l, int)
		if l == 1:
			x = np.array([-np.cos(math.pi / 2.)])
			w = np.array([2.])
		else:
			n = 2 ** (l - 1) + 1
			#n = 2*l - 1
			x = np.array([-np.cos( math.pi * i / (n - 1.)) for i in range(n)])
			w = np.zeros(n)
			w[0] = w[-1] = 1. / (n * (n - 2.))
			for i in range(1, n-1):
				s = [np.cos(2* math.pi * i * (j+1)/(n - 1.)) / (1. - 4*(j+1.)**2) for j in range((n-1)/2)]
				s[-1] = s[-1] / 2.
				w[i] = 2 * (1. + 2. * np.sum(s)) / (n - 1.)
		rule = {'x': x, 'w': w / 2.}
		return rule


	def RuleDiff(self, rule1, rule2, nested = 'NN'):
		#assert nested in ['NN', 'FN'], 'Rules have to be either Non-nested (NN) or Fully-nested (FN)'
		#if nested == 'NN':
		return np.hstack([rule1['x'], rule2['x']]), np.hstack([rule1['w'], -rule2['w']])
		#else:
	#		locs = [i for i in range(len(rule1[0])) for j in range(len(rule2[0])) if rule1[0][i]==rule2[0][j]]
		#	print locs
		#	rule1[1][locs] = rule1[1][locs] - rule2[1]
		#	return rule1[0], rule1[1]


	def get_rule(self, d, l, exp = True, custom = None):
		assert l > 0
		assert d > 0
		assert isinstance(l, int)
		assert isinstance(d, int)
		#rules = ['CC', 'GL', 'GH', 'NC']
		#assert rule in rules, 'Input rule not supported or non existing.'
		if d == 1:
			if self._rule == 'GH':
				grid = self.GaussHermite(l, exp)
				[x, w] = grid['x'], grid['w']
			elif self._rule == 'GL':
				grid = self.GaussLegendre(l, exp)
				[x, w] = grid['x'], grid['w']
			elif self._rule == 'GLag':
				grid = self.GaussLaguerre(l, exp)
				[x, w] = grid['x'], grid['w']
			elif self._rule == 'NC':
				grid = self.Trapezoidal(l)
				[x, w] = grid['x'], grid['w']
			elif self._rule == 'CC':
				grid = self.ClenshawCurtis(l)
				[x, w] = grid['x'], grid['w']
			else:
				print ('Custom Gauss Quadrature does not support exponential level growth. Using linear growth...')
				grid = self.Gauss(l, custom['pdf'], custom['supp'], False)
				[x, w] = grid['x'], grid['w']
			return x.reshape(x.shape[0], 1), w
		else:
			d0 = d - 1
			if self._rule == 'GH':
				Q1 = [self.GaussHermite(i, exp) for i in range(1,l+d0)]
				H_rule = self.GaussHermite(1, exp)
				D1 = [[H_rule['x'], H_rule['w']]]
			elif self._rule == 'GL':
				Q1 = [self.GaussLegendre(i, exp) for i in range(1,l+d0)]
				L_rule = self.GaussLegendre(1, exp)
				D1 = [[L_rule['x'], L_rule['w']]]
			elif self._rule == 'GLag':
				Q1 = [self.GaussLaguerre(i, exp) for i in range(1,l+d0)]
				L_rule = self.GaussLaguerre(1, exp)
				D1 = [[L_rule['x'], L_rule['w']]]
			elif self._rule == 'CC':
				Q1 = [self.ClenshawCurtis(i) for i in range(1,l+d0)]
				CC_rule = self.ClenshawCurtis(1)
				D1 = [[CC_rule['x'], CC_rule['w']]]
			elif self._rule == 'NC':
				Q1 = [self.Trapezoidal(i) for i in range(1,l+d0)]
				NC_rule = self.Trapezoidal(1)
				D1 = [[NC_rule['x'], NC_rule['w']]]
			elif self._rule == 'custom_pdf':
				print ('Custom Gauss Quadrature does not support exponential level growth. Using linear growth...')
				Q1 = [self.Gauss(i, custom['pdf'], custom['supp'], False) for i in range(1, l+d0)]
				G_rule = self.Gauss(1, custom['pdf'], custom['supp'], False)
				D1 = [[G_rule['x'], G_rule['w']]]
			for i in range(1,len(Q1)):
				D1 = D1 + [self.RuleDiff(Q1[i], Q1[i-1])]

			MI = self.mi_terms(l+d0-1, d0)
			THETA = np.zeros(d)
			W = np.zeros(1)
			for i in range(MI.shape[0]):
				mi = MI[i,:]
				k_abs = mi.sum()
				x = [(y,) for y in Q1[l+d0-1-k_abs]['x']]
				w = [(y,) for y in Q1[l+d0-1-k_abs]['w']]
				for j in range(d0):
					x = [u+(v,) for u in x for v in D1[mi[j]-1][0]]
					w = [u+(v,) for u in w for v in D1[mi[j]-1][1]]
				THETA = np.vstack([THETA, np.array(x)])
				W = np.hstack([W, np.prod(np.array(w), axis = 1)])
			#if self._rule in ['GH', 'GL', 'CC', 'NC'] and exp == True:
			
			if l > 1:
				W = np.delete(W, 0,0)
				[theta_uni, ind, inv, c] = np.unique(np.delete(THETA,0,0), True, True, True, axis = 0)
				locs = [j for j in range(c.shape[0]) if c[j] > 1]
				w_uni = W[ind]
				for j in locs:
					loc_j = [k for k in range(inv.shape[0]) if inv[k] == j]
					w_uni[j] = W[loc_j].sum()
				
				locs_0 = np.argwhere(np.abs(theta_uni.flatten()) < 1e-16)

				theta_flat = theta_uni.flatten()
				theta_flat[locs_0] = 0.
				return theta_flat.reshape(w_uni.shape[0],d), w_uni
			else:
				return np.delete(THETA, 0,0), np.delete(W, 0,0)





"""

"""


__all__ = ['PolyChaos', 'BasisAdapt']


try:
    import numpy as np
except ImportError:
    print ('numpy needs to be installed')

try:
    import scipy.stats as st 
    from scipy.linalg import polar
except ImportError:
    print ('scipy needs to be installed')

from _poly_basis import Hermite1d, PolyBasis


class PolyChaos(object):
    """
    Class for computing Polynomial Chaos coefficients
    and for generating corresponding samples
    """

    _name = None

    def __init__(self, name = 'PC expansion'):
        """
        Initializes the object
        """
        self._name = name


    def comp_coeffs(self, xi, y, w, degree = 1, poly_type = 'H'):
        """
        Computes Chaos coefficients by performing numerical integration 
        on the expression c_{alpha} = E[y psi_{alpha}(xi)] based on quadrature rules.
        Input
        :xi: array (N,d) where N the number of quadrature points and 
                               and d the dimension of the input variables.
        :y: array (N,) the model output values (should be of equal length with xi.shape[0]).
        :w: array (N,) the quadrature weights.
        :degree: int
                The highest order of the polynomials in the expansion.
        :poly_type: str 
               The type of polynomials in the expansion (currently supports only Hermite and Legendre)

        [1] Tipireddy, R. and Ghanem, R., 2014. Basis adaptation in homogeneous chaos spaces. 
        Journal of Computational Physics, 259, pp.304-317.
        """
        assert xi.shape[0] == y.shape[0]
        assert xi.shape[0] == w.shape[0]
        assert isinstance(degree, int)
        d = xi.shape[1]
        if poly_type not in ['H', 'L']:
            raise ValueError('Unknown type of polynomials !')
        pol = PolyBasis(d, degree, poly_type)
        return np.dot(y * w, pol(xi))


    def sample(self, coeffs, dim, degree, N_mc = 1e+3, poly_type = 'H'):
        """  
        Generates samples from a Polynomial Chaos expansion whose coefficients and type 
        of polynomials are provided.
        """
        assert N_mc > 0
        assert isinstance(N_mc, int)

        pol = PolyBasis(dim , degree)
        assert pol._MI_terms.shape[0] == coeffs.shape[0]

        if poly_type == 'H':
            X = st.norm.rvs(size = (N_mc, dim))
        elif poly_type == 'L':
            X = st.uniform(loc = -1., scale = 2.).rvs(size = (N_mc, dim))
        else:
            raise ValueError('Unknown type of polynomials !')
        return np.dot(pol(X), coeffs)


    def sample_many_QoI(self, coeffs, orders, dim, MC1 = 1e5, X1 = [213], ptype = 'Hermite', seed = 1):
        """
        Generates Monte Carlo samples for multiple QoIs for which the PCE coefficients are provided. Allows 
        PCEs to have different highest order.

        Input: 
                coeffs: list that contains numpy arrays with shape (NC_i,) where NC_i the number of coeffs for the ith QoI.
                orders: list that contains integers specifying the order of the ith QoI.
                dim : int, the input dimensionality
                The rest as in generate_samples_from_pce
        """
        assert len(coeffs) == len(orders)
        m = len(coeffs)
        samples = np.zeros((MC1,m))
        for i in range(m):
            [samples[:,i], xi] = self.generate_samples_from_pce(self, coeffs[i],dim=dim,order=orders[i],MC1=MC1,X1=X1,ptype=ptype, seed = seed)
        return np.vstack([xi, samples])



    def l2_error(self, q_1, q_2, d1, d2, order):
        """ l2-norm relative error function
        Input
        :q_1: vector of coefficients corresponding to eta_{i-1}
        :q_2: vector of coefficients corresonding to eta_{i-2}
        Return
        :e: relative l2-error ||q_1 - q_2|| / ||q_2||   

        """
        import math
        assert np.shape(q_2)[0] == math.factorial(d2+order) / (math.factorial(order) * math.factorial(d2))
        assert np.shape(q_1)[0] < np.shape(q_2)[0]
        Q1 = np.zeros(q_2.shape[0])
        #mi = PolyBasis().mi_terms(order, d2)
        Q1[PolyBasis().mi_terms_loc(d1, d2, order)] = q_1
        return (np.linalg.norm((Q1 - q_2),2) / np.linalg.norm(q_2,2))


    def sobol_ind(self, idx, od, dim, C):      
        """ Computate PC-based 
        Sobol sensitivity
        indices 
        
        Input
        :idx: index of xi for which to compute
              Sobol's sensitivity indices 
              (start from 0)
        :od: order of highest polynomial of PCE
        :dim: dimensionality
        :C: matrix of PCE coefficients
            corresponding to the multi-
            index matrix
        
        Return 
        :su: Sobol PC-based sensitivity indices 
             corresponding to xi's at indices idx

        Ref: 
            Sudret (2006) 'Global sensitivity analysis
            using polynomial chaos expansion'  
        """
        assert isinstance(idx, list)
        if len(C.shape) == 1:
            C = C.reshape(C.shape[0],1)

        MI = PolyBasis().mi_terms(dim, od)
        assert MI.shape[0] == C.shape[0]
        loc = range(MI.shape[0])

        for i in np.arange(dim):
            ai = None
            if i in idx:
                ai = np.where(MI[:,i] > 0)
            else:
                ai = np.where(MI[:,i] == 0)
            loc = np.intersect1d(loc,ai)
            
        Cx = C[loc,:]  
        su = np.zeros(Cx.shape[1])
        
        for m in range(np.shape(Cx)[1]):
            su[m] = np.sum(Cx[:,m]**2) / np.sum(C[1:,m]**2)
        
        return su


class BasisAdapt(PolyChaos):
    """ Basis adaptation """


    def __init__(self, name = 'Adapted PC expansion'):
        """
        Ininialize the object
        """

        super(BasisAdapt, self).__init__(name = name)


    def gauss_adaptation(self, coeffs, dim, method = 0):
        """
        Computes the rotation matrix A that corresponds to the Gaussian adaptation [1]. 
        Input
        :coeffs : array 
            The coefficients of the first order Hermite polynomials (linear part)
            that consist the first row of A.
        :dim: int 
            Dimension of the original input germs xi.
        :method: int {0,1,2,3}
            The method to compute the isometry. 
                0 (default) : Via orthogonal decomposition of a * a.T
                1 : Via a Gram-Schmidt procedure on the matrix C with coeffs (normalized) at its first row,
                        ones along its diagonal, zeros anywhere else.
                2 : Via orthogonal decomposition of the Householder matrix H = I - 2 u u' / ||u|| ** 2
                3 : Via a Gram-Schmidt procedure on the matrix C with coeffs (normalized) at its first row,
                        and the ith largest coeff on the column corresponding to its variable xi
                        on the (i+1) row, zero elsewhere
                4 : Via a Gram-Schmidt procedure on the matrix C with N-QoI coeffs (normalized) at its 
                        first N rows, ones along its diagonal, zeros anywhere else.
                5 : Via a polar decomposition of matrix C with coeffs (normalized) at its first row,
                        and the ith largest coeff on the column corresponding to its variable xi
                        on the (i+1) row, zero elsewhere 
        [1] Tipireddy, R. and Ghanem, R., 2014. Basis adaptation in homogeneous chaos spaces. 
        Journal of Computational Physics, 259, pp.304-317.
        """

        assert coeffs.shape[0] == dim
        if method != 4:
            a = coeffs.reshape((coeffs.shape[0], 1))
        elif method == 4:
            a = coeffs.reshape((coeffs.shape[0], coeffs.shape[1]))
        else:
            print ('Shape of first order coefficients unrecognized')

        #H = np.eye(dim) - 2 * np.dot(u, u.T) / np.linalg.norm(u) ** 2 # Householder matrix
        if method == 0:
            C = np.dot(a, a.T)
            [vals, vecs] = np.linalg.eigh(C)
            return vecs[:,::-1].T
        elif method == 1: # Gram-Schmidt
            C = np.eye(a.shape[0])
            C[0,:] = a[:,0]
            [q,r] = np.linalg.qr(C.T)
            return q.T
        elif method == 2: # Householder matrix
            H = np.eye(dim) - 2 * np.dot(a, a.T) / np.linalg.norm(a) ** 2 # Householder matrix
            [vals, vecs] = np.linalg.eigh(H)
            return np.real(vecs).T
        elif method == 3:
            c3 = np.argsort(np.abs(coeffs))[::-1]
            C = np.zeros((dim, dim))
            C[:,0] = a[:,0]
            j = 0        
            for i in range(0, np.size(coeffs)-1):
                C[c3[j], i+1] = coeffs[c3[j]]
                j += 1
            [q,r] = np.linalg.qr(C)
            q=np.mat(q)
            return q.T
        elif method == 4: # Multi-QoI with Gram-Schmidt
            C = np.eye(a.shape[0])
            for k in np.arange(coeffs.shape[1]):
                C[k,:] = a[:,k]    
                [q,r] = np.linalg.qr(C.T)
            return q.T
        elif method == 5: # Polar decomposition
            c3 = np.argsort(np.abs(coeffs))[::-1]
            C = np.zeros((dim, dim))
            C[:,0] = a[:,0]
            j = 0        
            for i in range(0, np.size(coeffs)-1):
                C[c3[j], i+1] = coeffs[c3[j]]
                j += 1
            [u,p] = polar(C)
            return u.T, np.linalg.det(C)            
        else:
            raise ValueError('Method parameter must be in {0,1,2,3,4,5}')

    def quadratic_adaptation(self, coeffs, dim):
        """
        Computes what is referred to as Quadratic Adaptation in [1], using chaos coefficients of the 
        2nd ordre Hermite polynomials.
        Input
        :coeffs: array
                The coefficients of the second order Hermite polynomials (quadratic part)
        :dim: int
                Dimension of the input germs xi.

        [1] Tipireddy, R. and Ghanem, R., 2014. Basis adaptation in homogeneous chaos spaces. 
        Journal of Computational Physics, 259, pp.304-317.  
        """

        assert coeffs.shape[0] == dim * (dim + 1) / 2
        Q_upper = np.zeros((dim, dim))
        for l in range(dim):
            for k in range(dim-l):
                if k == 0:
                    Q_upper[l,l+k] = coeffs[np.sum(np.arange(dim)[dim-l:]) + k] / np.sqrt(2.)
                else:
                    Q_upper[l,l+k] = coeffs[np.sum(np.arange(dim)[dim-l:]) + k] 
        S = Q_upper + np.transpose(Q_upper) - np.diag(np.diag(Q_upper))
        [vals, vecs] = np.linalg.eigh(S, UPLO = 'U')
        return vecs.T

    def eta_to_xi_mapping(self, eta, A, zeta = None):
        """
        Maps points from the (lower dimensional) eta-space to the xi space. The dimensionality of eta
        must be smaller than or equal to that of xi (shape of A) and the rotation matrix must be a 
        unitaty matrix.
        Parameters ---- > eta : array (N, d0) where N the number of points, d0 the dimensionality of eta
                                The set of N (quadrature) points on a d0-dimensional space.
                            A : array (d, d)
                                Isometry which serves as the rotation matrix (xi = A' [eta, zeta]) 
                         zeta : array (N, d-d0) 
                                Optional set of N points on a (d-d0)-dimensional space that will 
                                augment eta to the d-dimensional space. Zero values will be used at the 
                                default choice when not user-specified.

        [1] Tipireddy, R. and Ghanem, R., 2014. Basis adaptation in homogeneous chaos spaces. 
        Journal of Computational Physics, 259, pp.304-317.
        """

        assert A.shape[0] == A.shape[1]

        d0 = eta.shape[1]
        d = A.shape[0]
        N = eta.shape[0]
        if zeta == None:
            zeta = np.zeros((N, d-d0))
        else:
            assert eta.shape[0] == zeta.shape[0]
            assert eta.shape[1] + zeta.shape[1] == A.shape[0]
        eta_full = np.hstack([eta, zeta]) # Augment eta points with zeta
        return np.dot(A.T, eta_full.T).T


    def transform_coeffs(self, coeffs, deg, iso, eta_dim = 1, method = 0, num_MC = 10000, xi = None, w = None):
        """
        Given the coefficients of a (low dimensional) chaos expansion with respect to the 
        eta basis, where eta = A * xi, transforms the coefficients to those that correspond 
        to an expansion with respect to the original xi basis.
        Input
        :coeffs: array(n,) 
                The coefficients of the PCE wrt the eta basis
        :iso: array(d,d) where d is the dimensionality of xi.
               The (unitary) matrix that relates eta = A * xi.
        :eta_dim: int
                The number of eta components used in the current PCE (eta_dim <= d).
        :num_MC: int
                The number of Monte Carlo samples to be used for estimating the 
                inner products <psi_{beta}(A * xi), psi_{alpha}(xi)>

        [1] Tipireddy, R. and Ghanem, R., 2014. Basis adaptation in homogeneous chaos spaces. 
        Journal of Computational Physics, 259, pp.304-317.
        """

        assert isinstance(eta_dim, int)
        assert isinstance(deg, int)
        assert eta_dim <= iso.shape[0]
        pol_eta = PolyBasis(eta_dim, deg)
        assert len(pol_eta._terms) == coeffs.shape[0]
        if method == 0:
            xi = st.norm.rvs(size = (num_MC, iso.shape[1]))
            eta = np.dot(iso, xi.T)[:eta_dim, :].T
            pol = PolyBasis(iso.shape[0], deg)
            return np.dot(coeffs, np.dot(pol_eta(eta).T, pol(xi))) / num_MC
        elif method == 1:
            eta = np.dot(iso, xi.T)[:eta_dim, :].T
            pol = PolyBasis(iso.shape[0], deg)
            return np.dot(coeffs, np.dot(pol_eta(eta).T * w, pol(xi)))
        else:
            raise ValueError('For integration with Monte Carlo or quadrature rule choose 0 or 1 respectively.')






"""
Class that generates quadrature points corresponding to sparse grids.
Requirement: Dakota (https://dakota.sandia.gov/)

Author: Panagiotis Tsilifis
Date: 2/10/2017
"""

__all__ = ['SparseGrid']

from os import system
import numpy as np

class SparseGrid(object):
	"""

	"""

	# Name of the SparseGrid object.
	__name__ = None

	# The level of the quadrature rule.
	_level = None 

	# The type of variables.
	_variables = None

	# The parameters of the variables. 
	_params = None

	@property
	def level(self):
		return self._level

	@level.setter
	def level(self, value):
		self._level = value

	@property
	def variables(self):
		return self._type

	@variables.setter
	def variables(self, value):
		self._type = value

	@property
	def params(self):
		return self._params

	@params.setter
	def params(self, value):
		assert value.shape[0] == 2
		if self._variables == "N" or "Normal":
			assert value[1] > 0
			self._params = value
		else:
			self._params = value

	def __init__(self, level, variables, params = None, name = "Sparse Grid Generator"):

		assert isinstance(level, int)
		self._level = level
		if variables == "N" or "Normal":
			self._variables = variables
			if params is not None:
				assert params.shape[0] == 2
				assert params[1] > 0
				self._params = params
			else:
				self._params = np.array([0., 1.])
		elif variables == "U" or "Uniform":
			self._variables = variables
			if params is not None:
				assert params.shape[0] == 2
				self._params = params
			else:
				self._params = np.array([-1., 1.])
		else:
			raise RuntimeError('Currently only Normal and Uniform grids are supported !\n Type should be "Normal" ("N") or "Uniform" ("U").')
		self.__name__ = str(name)

	def generate_grid(self, dim):
		"""
		Generates a sparse grid of dimension "dim" with default parameters. 
		"""
		assert isinstance(dim, int)
		name = 'sparse_grid_gen.in'
		f = open(name, 'w')
		if self._variables == "N" or "Normal":
			text = ['method\n', '	polynomial_chaos\n', '	sparse_grid_level=3\n', '	output verbose\n', 
					'variables\n', '	normal_uncertain ='+str(dim)+'\n', '	means ='+str(dim)+'*'+str(self._params[0])+'\n', '	std_deviations ='+str(dim)+'*'+str(self._params[1])+'\n',
					'interface\n', '	direct\n', "	analysis_driver = 'rosenbrock'\n", 'responses\n', 
					"	response_functions = 1\n", "	no_gradients\n", "	no_hessians\n"]
		elif self._variables == "U" or "Uniform":
			text = ['method\n', '	polynomial_chaos\n', '	sparse_grid_level=3\n', '	output verbose\n', 
					'variables\n', '	uniform_uncertain ='+str(dim)+'\n', '	lower_bounds ='+str(dim)+'* '+str(self._params[0])+'\n', '	upper_bounds ='+str(dim)+'*'+str(self._params[1])+'\n',
					'interface\n', '	direct\n', "	analysis_driver = 'rosenbrock'\n", 'responses\n', 
					"	response_functions = 1\n", "	no_gradients\n", "	no_hessians\n"]

		f.writelines(text)
		f.close()
		system('dakota -i ' + str(f.name))

		out = open('dakota_sparse_tabular.dat', 'rU')
		N = -1
		for line in f:
			N = N + 1
		weights = np.zeros(N)
		grid = np.zeros((N, dim))
		f.readline()

		pos = []
		for i in range(dim):
			pos += [25+i*25]

		for i in range(N):
			line = f.readline()
			print line[:6]
			weights[i] = float(line[6:25])
			grid[i,:] = np.array([float(line[j:j+15]) for j in pos])

		return grid, weights
			

	def __str__(self):
		"""
		Return a string representation of the object.
		"""
		s = 'Name: ' + self.__name__ + '\n'
		s += 'Level: ' + str(self._level) + '\n'
		s += 'Variables: ' + str(self._variables) + '\n'
		return s





"""
Implements a class representing an adapted Polynomial Chaos expansion.

Author : Panos Tsilifis
Date : 2/7/2017

"""

__all__ = ['BasisAdaptation']


import numpy as np 


class BasisAdaptation(object):
	"""
	A class that represents a Polynomial Chaos expansion 
	expressed in terms of a rotated basis. 

	"""

	# A name for the expansion
	__name__ = None

	# The type of polynomials in the expansion
	_poly_type = None

	# The initial input dimension.
	_inp_dim = None

	# The initial highest order of the Chaos space.
	_chaos_order = None

	# The series coefficients with respect to the initial input xi.
	_chaos_coeffs = None

	# The number of coefficients with respect to the initial input xi.
	_num_chaos_coeffs = None

	@property
	def inp_dim(self):
		"""
		:getter: The initial input dimension
		"""
		return self._inp_dim

	@inp_dim.setter
	def inp_dim(self, value):
		"""
		Set the input dimension.
		"""
		self._inp_dim = value

	@property
	def chaos_order(self):
		"""
		:getter: The initial highest order of the Chaos space.
		"""
		return self._chaos_order

	@chaos_order.setter
	def chaos_order(self, value):
		"""
		Set the initial highest order of the Chaos space.
		"""
		self._chaos_order = value

	@property
	def num_chaos_coeffs(self):
		"""
		:getter: The number of coefficients with respect to the initial input xi.
		"""
		return self._num_chaos_coeffs

	@num_chaos_coeffs.setter
	def num_chaos_coeffs(self, value):
		"""
		Set the number of coefficients with respect to the initial input xi.
		"""
		self._num_chaos_coeffs = value

	@property
	def poly_type(self):
		"""
		:getter: The type of polynomials in the expansion 
		(currently supports only Hermite and Legendre).
		"""
		return self._poly_type

	@poly_type.setter
	def poly_type(self, value):
		"""
		Set the type of polynomials to be either Hermite of Legendre.
		"""
		if value == 'Hermite' or "H" or 'Legendre' or "L":
			self._poly_type = value
		else:
			raise RuntimeError('The polynomials should be either Hermite of Legendre!')

	def __init__(self, num_dim, num_chaos_coeffs = None, chaos_order = None, chaos_coeffs = None, pol_type = 'Hermite', name = 'Adapted PC expansion'):
		"""
		Initialize the object.
		"""
		assert isinstance(num_dim, int)
		self._inp_dim = num_dim
		if num_chaos_coeffs is not None:
			self._num_chaos_coeffs = num_chaos_coeffs
		if chaos_order is not None:
			self._chaos_order = chaos_order
		if chaos_coeffs is not None:
			self._chaos_coeffs = chaos_coeffs
		if pol_type == 'Hermite' or 'Legendre' or 'H' or 'L':
			self._poly_type = pol_type
		else:
			raise RuntimeError('The polynomials should be either Hermite of Legendre!')
		self.__name__ = str(name)


	def __str__(self):
		"""
		Return a string representation of the object.
		"""
		s = 'Name: ' + self.__name__ + '\n'
		s += 'Type of polynomials: ' + self._poly_type + '\n'
		s += 'Initial input dimension: ' + str(self._inp_dim) + '\n'
		s += 'Order of expansion: ' + str(self._chaos_order) + '\n'
		return s











"""
Implements a class for performing Basis Adaptation using the Active Subspace method.

Author: Panos Tsilifis
Date: 2/7/2017

"""

__all__ = ['ActiveSubspaceAdaptation', 'GaussianAdaptation', 'QuadraticAdaptation']


import numpy as np
import scipy.stats as st
from . import PolyBasis
from . import MonicPoly
from . import Hermite1d
from . import Legendre1d
from . import Laguerre1d
from . import BasisAdaptation
from . import dirac1D
from . import diracND
from . import grad_inner_prod_Legendre
from . import gradgrad_inner_prod_Legendre

class ActiveSubspaceAdaptation(BasisAdaptation):
    """
    A class the represents a Polynomial Chaos expansion with adapted basis using the 
    Active Subspace method. 

    """

    def __init__(self, num_dim, pol_type = 'H', name = 'Active Subspace Basis Adaptation'):
        """
        Initialize the object
        """
        super(ActiveSubspaceAdaptation, self).__init__(num_dim, pol_type = pol_type, name = name)

    def _stiffness_K(self, deg, i, j):
        """
        Computes the stiffness matrix K^{ij} with entries 
        K_{ab}^{ij} = E[\frac{d psi_a (xi)}{d xi_i} \frac{d psi_b(xi)}{d xi_j}]
        """
        assert isinstance(deg, int)
        assert isinstance(i, int)
        assert isinstance(j, int)
    	assert i < self._inp_dim and j < self._inp_dim
        #rvs = [st.norm()] * self._inp_dim
        
        #pol = orthpol.ProductBasis(rvs, degree = deg)
        pol = PolyBasis(self._inp_dim, deg)
        terms = pol.mi_terms(self._inp_dim, deg)
        Q = terms.shape[0]
        stiff = np.zeros((Q,Q))
        if self._poly_type == 'Legendre' or 'L':
            for k in range(Q):
                for l in range(Q):
                    alpha = terms[k,:]
                    beta = terms[l,:]
                    if i == j :
                        a_m = np.delete(alpha, i)
                        b_m = np.delete(beta, i)
                        a_i = alpha[i]
                        b_i = beta[i]
                        E_ii = gradgrad_inner_prod_Legendre(a_i,b_i)
                        stiff[k,l] = diracND(a_m, b_m) * E_ii
                    else:
                        a_m = np.delete(alpha, [i,j])
                        b_m = np.delete(beta, [i,j])
                        a_i = alpha[i]
                        a_j = alpha[j]
                        b_i = beta[i]
                        b_j = beta[j]
                        E_i = grad_inner_prod_Legendre(a_i, b_i)
                        E_j = grad_inner_prod_Legendre(b_j, a_j)
                        stiff[k,l] = diracND(a_m, b_m) * E_i * E_j
        elif self._poly_type == 'Hermite' or 'H':
            for k in range(Q):
                for l in range(Q):
                    alpha = terms[k,:]
                    beta = terms[l,:]
                    if alpha[i] - 1 < 0 or beta[j] - 1 < 0:
                        stiff[k,l] = 0
                    else:
                        C = np.sqrt(alpha[i] * beta[j])
                        alpha[i] = alpha[i] - 1
                        beta[j] = beta[j] - 1
                        stiff[k,l] = C * diracND(alpha, beta)
        return stiff

    def _grad_covar(self, deg, coeffs):
        """
        Computes the covariance matric of the gradient vector 
        of the QoI given a set of coefficient values "coeffs"
        corresponding to the currently availble full dimensional
        PCE of the QoI.
        """
        rvs = None
        #if self._poly_type == 'Hermite' or 'H':
        #    rvs = [st.norm()] * self._inp_dim
        pol = PolyBasis(self._inp_dim, deg, self._poly_type[0])
        #elif self._poly_type == 'Legendre' or 'L':
        #    rvs = [st.uniform()] * self._inp_dim
        #pol = orthpol.ProductBasis(rvs, degree = deg)
        terms = pol.mi_terms(self._inp_dim, deg)
        Q = terms.shape[0]
        assert Q == coeffs.shape[0]
        Grad_lo = np.zeros((self._inp_dim, self._inp_dim))
        for i in range(self._inp_dim):
            for j in range(i+1):
                stiff = self._stiffness_K(deg, i, j)
                Grad_lo[i,j] = np.dot(np.dot(coeffs.reshape(1,Q), stiff), coeffs.reshape(Q,1))[0,0]
        return Grad_lo + Grad_lo.T - np.diag(np.diag(Grad_lo))

    def rotation(self, coeffs):
        """
        Computes the rotation matrix.
        """
        raise NotImplementedError('Only my children compute the rotation matrix directly!')


class GaussianAdaptation(ActiveSubspaceAdaptation):
    """
    A class that represents a Polynomial Chaos expansion with adapted basis using the 
    Gaussian adaptation method. 
    """
    def __init__(self, num_dim, name = 'Gaussian Basis Adaptation'):
        super(GaussianAdaptation, self).__init__(num_dim, pol_type = 'H', name = name)

    def rotation(self, coeffs):
        assert coeffs.shape[0] > self._inp_dim
        coeffs = coeffs[:self._inp_dim + 1]
        C = self._grad_covar(1, coeffs)
        [l, v] = np.linalg.eigh(C)
        return v[:,::-1].T


class QuadraticAdaptation(ActiveSubspaceAdaptation):
    """
    A class that represents a Polynomial Chaos expansion with adapted basis using the 
    Quadratic adaptation method. 
    """
    def __init__(self, num_dim, name = 'Quadratic Basis Adaptation'):
        super(QuadraticAdaptation, self).__init__(num_dim, pol_type = 'H', name = name)

    def rotation(self, coeffs):
        Q = self._inp_dim + 1 + self._inp_dim * (self._inp_dim + 1) / 2
        assert coeffs.shape[0] >= Q
        coeffs = coeffs[:Q]
        coeffs[:self._inp_dim + 1] = 0
        C = self._grad_covar(2, coeffs)
        [l, v] = np.linalg.eigh(C)
        return v[:,::-1].T
        



"""
Initialize the Basis Adaptation Python module.

Author : Panos Tsilifis
Date : 2/7/2017

"""

from _tools import *
from _poly_basis import *
from _poly_chaos import *
from _basis_adaptation import *
from _AS_adaptation import *
from _quadrature import *
from _sparse_grid import *



from setuptools import setup

setup(name='chaos_basispy',
	version = '0.2.1',
	description = 'A package for Polynomial Chaos basis rotation',
	long_description = 'The package chaos_basispy defines the chaos_basispy module which attempts '+\
						'to collect the recently developed Basis Adaptation (BA) techniques that ' +\
						'have been developed particularly for Polynomial Chaos expansions (PCE).'+\
						'The core idea behind Basis Adaptation is to apply a transformation on the '+\
						'input variables in a way such that the Quantity of Interest (QoI) which we '+\
						'are approximating via a PCE can be expressed with respect to a reduced basis. '+\
						'Provided that the input variables are Gaussian and the transformation is through '+\
						'an isometry (unitary matrix), the new variables preserve "Gaussianity" and a new PCE '+\
						'can be constructed with respect to them. This makes the BA framework particularly '+\
						'attractive on Hermite PCEs and in fact inapplicable (yet) to generalized PCEs. An '+\
						'expeption is when the new variables is 1-dimensional which can be mapped to a uniform '+\
						'r.v. through its own cdf and therefore the Legendre adapted PCE can be constructed. Due '+\
						'to the above, this module considers only Hermite and Legendre Chaos expansions.',
	url = 'https://github.com/tsilifis/chaos_basispy',
	author = 'Panagiotis Tsilifis',
	author_email = 'panagiotis.tsilifis@epfl.ch',
	keywords = ['polynomial chaos', 'hermite chaos', 'legendre_chaos', 'dimensionality reduction', 'surrogate', 'uncertainty quantification'],
	packages = ['chaos_basispy'])



"""
Various tool functions used throughout the package.

Author: Panos Tsilifis
Date: 2/7/2017

"""


__all__ = ['dirac1D', 'diracND', 'grad_inner_prod_Legendre', 'gradgrad_inner_prod_Legendre', 'inverse_transform_sampling']


import numpy as np
import scipy.stats as st
from scipy import interpolate

def dirac1D(a,b):
    """
    Univariate Dirac function.
    """
    val = 0 
    if a == b:
        val = 1
    return val

def diracND(a,b):
    """
    Multivariate Dirac function.
    """
    assert len(a) == len(b)
    n = len(a)
    val = 1
    for i in range(n):
        val = val * dirac1D(a[i], b[i])
    return val

def grad_inner_prod_Legendre(a,b):
    """
    Computes the inner product (expectation) E[psi'_a(xi) psi_b(xi)],
    where psi'_a(xi) the derivative of a Legendre polynomial of order a
    and psi_b(xi) a Legendre polynomial of order b. Both polynomials as assumed
    normalized and xi follows Uniform(-1,1).
    """
    a = int(a)
    b = int(b)
    val = 0
    if a == 1:
        val = np.sqrt(3) * dirac1D(0,b)
    elif a > 1:
        c =  a - 1
        while c > 0:
            val = val + dirac1D(c, b)
            c = c - 2
        val = val * np.sqrt((2*a+1)*(2*b+1))
    return val


def gradgrad_inner_prod_Legendre(a,b):
    """
    Computes the inner product (expectation) E[psi'_a(xi) psi'_b(xi)],
    where psi'_a(xi) the derivative of a Legendre polynomial of order a
    and psi'_b(xi) the derivative of a Legendre polynomial of order b. Both 
    polynomials as assumed normalized and xi follows Uniform(-1,1).
    """
    a = int(a)
    b = int(b)
    val = 0.
    if a == 1 and b > 0:
        if b == 1:
            val = 3.
        elif b > 1 and b%2 == 0:
            val = 0.
        elif b > 1 and b%2 != 0:
            val = val + np.sqrt(3*(2*b+1))
    elif a > 1 and b > 0:
        c = range(a-1, 0, -2)
        d = range(b-1, 0, -2)
        for i in range(len(c)):
            for j in range(len(d)):
                val = val + (2*c[i]+1) * dirac1D(c[i], d[j])
        val = val * np.sqrt((2*a+1)*(2*b+1))
    return val


def inverse_transform_sampling(data, eval_points = 'None', n_bins=40, n_samples=1000):
    hist, bin_edges = np.histogram(data, bins=n_bins, density=True)
    cum_values = np.zeros(bin_edges.shape)
    cum_values[1:] = np.cumsum(hist*np.diff(bin_edges))
    cum_values[-1] += 1.e-15
    inv_cdf = interpolate.interp1d(cum_values, bin_edges)
    if eval_points == 'None':
        eval_points = np.random.rand(n_samples)
    return inv_cdf(eval_points)



""" Basis terms Module 

Author: Panos Tsilifis
Date: 7/10/2017

"""

__all__ = ['PolyBasis', 'MonicPoly', 'Hermite1d', 'Legendre1d', 'Laguerre1d']

import numpy as np
import math
from scipy import misc
import itertools as itls


class Hermite1d(object):
    """
    Class that constructs 1-dimensional Hermite polynomials.
    """


    _degree = None


    def __init__(self, degree = 1):
        """
        Ininializes the object 
        """
        assert isinstance(degree, int)
        assert degree > 0
        self._degree = degree 

    def eval(self, x):
        H = np.zeros(self._degree + 1)
        H[0] = 1.
        H[1] = x
        for i in range(2,H.shape[0]):
            H[i] = (x * H[i-1] - (i-1) * H[i-2] )
        H = H / [math.sqrt(math.factorial(i)) for i in range(H.shape[0])]
        return H

    def __call__(self, x):
        N = x.shape[0]
        H = np.zeros((N, self._degree + 1))
        for i in range(N):
            H[i,:] = self.eval(x[i])
        return H


class Legendre1d(object):
    """
    Class that contains 1-dimensional Legendre polynomials.
    """

    _degree = None

    def __init__(self, degree = 1):
        """
        Initializes the object
        """
        assert isinstance(degree, int)
        assert degree > 0
        self._degree = degree

    def eval(self, x):
        H = np.zeros(self._degree + 1)
        H[0] = 1.
        H[1] = x
        for i in range(2, H.shape[0]):
            H[i] = ( (2.*i-1.)* x * H[i-1] - (i-1.)* H[i-2] ) / i
        H = H / [math.sqrt(1. / (2.*i+1.)) for i in range(H.shape[0])]
        return H

    def __call__(self, x):
        N = x.shape[0]
        H = np.zeros((N, self._degree + 1))
        for i in range(N):
            H[i,:] = self.eval(x[i])
        return H


class Laguerre1d(object):
    """
    Class that contains 1-dimensional Laguerre polynomials.
    """

    _degree = None

    def __init__(self, degree = 1):
        """
        Initializes the object
        """
        assert isinstance(degree, int)
        assert degree > 0
        self._degree = degree

    def eval(self, x):
        H = np.zeros(self._degree + 1)
        H[0] = 1.
        H[1] = 1. - x
        for i in range(2, H.shape[0]):
            H[i] = ((2*(i-1)+1-x)*H[i-1] - (i-1)*H[i-2]) / (i)
        return H

    def __call__(self, x):
        N = x.shape[0]
        H = np.zeros((N, self._degree + 1))
        for i in range(N):
            H[i,:] = self.eval(x[i])
        return H


class MonicPoly(object):
    """
    Class that contains 1-dimensioanl monic polynomials, 
    orthogonal with respect to an arbitrary measure. 
    """

    _degree = None
    _measure = None
    _support = None

    def __init__(self, degree, pdf, supp):
        """
        Initializes the object. 
        """
        assert isinstance(degree, int)
        assert degree > 0
        assert len(supp) == 2
        assert isinstance(supp, list)
        self._degree = degree
        self._measure = pdf
        self._support = supp


    def gram_schmidt(self, grid = 1000):
        """
        Constructs polynomials of order 
        up to "order" that are orthogonal with
        respect to the arbitrary probability 
        measure provided by "ker".
        """
        monos = [lambda x: x**0]
        for i in range(1, self._degree + 1):
            monos = monos + [lambda x: x**i]
        c = np.zeros((self._degree, self._degree))
        dom = np.linspace(self._support[0], self._support[1], grid)
        h = dom[1] - dom[0]
        polys = np.zeros((self._degree+1, grid))
        polys[0,:] = monos[0](dom)
        for i in range(1, self._degree+1):
            for j in range(i):
                c[i-1,j] = np.sum(monos[i](dom) * polys[j,:] * self._measure(dom)) / np.sum(polys[j,:] ** 2 * self._measure(dom))
            polys[i, :] = monos[i](dom) - np.dot(c[i-1,:i], polys[:i,:]) 
        return polys, c


    def recurr_coeffs(self, grid = 1000):
        t = lambda x: x
        poly, c = self.gram_schmidt(grid)
        dom = np.linspace(self._support[0], self._support[1], grid)
        h = dom[1] - dom[0]
        alphas = np.zeros(poly.shape[0])
        betas = np.zeros(poly.shape[0]-1)
        alphas[0] = np.sum(t(dom) * self._measure(dom)) * h
        for i in range(poly.shape[0]-1):
            alphas[i+1] = np.sum( t(dom) * poly[i+1,:]**2 * self._measure(dom)) * h / (np.sum(poly[i+1,:]**2 * self._measure(dom)) * h)
            betas[i] = np.sum(poly[i+1,:]**2 * self._measure(dom)) * h / (np.sum(poly[i,:] ** 2 * self._measure(dom)) * h)
        return alphas, betas


class PolyBasis(object):
    """ construct basis terms matrix """

    _degree = None 
    _dim = None
    _MI_terms = None
    _type = None
    _trunc = None

    def __init__(self, dim = 1, degree = 1, pol_type = 'H', trunc = 'TD', q = None):
        """
        Ininializes the object 
        """
        assert isinstance(dim, int)
        assert isinstance(degree, int)
        assert dim > 0
        assert degree > 0
        assert pol_type in ['H', 'L', 'Lag'], 'Only Hermite, Legendre and Laguerre polynomials are currently supported ! Choose among ' + str(['H', 'L', 'Lag']) + ' !'
        assert trunc in ['TD', 'TP', 'LQ'], 'Only total degree (TD), tensor product (TP) and L_q truncation is currently supported ! Hyperbolic cross truncation is also underway !'
        
        self._degree = degree 
        self._dim = dim
        if trunc == 'LQ' and q is None:
            q = 0.5
        self._MI_terms = self.mi_terms(self._dim, self._degree, trunc, q)
        self._type = pol_type
        self._trunc = trunc


    def __call__(self, XI):
        assert XI.shape[1] == self._dim
        if self._type == 'H':
            H = [Hermite1d(degree = self._degree)(XI[:,i]) for i in range(self._dim)]
            PSI = np.ones((XI.shape[0], self._MI_terms.shape[0]))
            for i in range(self._MI_terms.shape[0]):
                for j in range(self._dim):
                    PSI[:,i] *= H[j][:,self._MI_terms[i,j]]
        elif self._type == 'L':
            H = [Legendre1d(degree = self._degree)(XI[:,i]) for i in range(self._dim)]
            PSI = np.ones((XI.shape[0], self._MI_terms.shape[0]))
            for i in range(self._MI_terms.shape[0]):
                for j in range(self._dim):
                    PSI[:,i] *= H[j][:,self._MI_terms[i,j]]
        else:
            H = [Laguerre1d(degree = self._degree)(XI[:,i]) for i in range(self._dim)]
            PSI = np.ones((XI.shape[0], self._MI_terms.shape[0]))
            for i in range(self._MI_terms.shape[0]):
                for j in range(self._dim):
                    PSI[:,i] *= H[j][:,self._MI_terms[i,j]]
        return PSI


    def mi_terms(self, dim = None, order = None, trunc = 'TD', q = None):
        """ matrix of multi-indices corresponding to basis terms
        Input
        :order: PCE order
        :dim: PCE dimension
        :trunc: truncation type 
            'TD' --> total degree (sum alpha_i <= order)
            'TP' --> tensor product (max alpha_i <= order)
            'LQ' --> l_q truncation (sum alpha_i^q <= order^q)
            'HC' --> hyperbolic cross (prod [alpha_i+1] <= [order + 1])
        """
        assert trunc in ['TD', 'TP', 'LQ'], 'Only total degree (TD), tensor product (TP) and L_q truncation is currently supported ! Hyperbolic cross truncation is also underway !'
        
        if dim is None:
            dim = self._dim
        if order is None:
            order = self._degree

        def TD(dim, order):
            q_num = [int(misc.comb(dim+i-1, i)) for i in range(order+1)]
            mul_ind = np.array(np.zeros(dim, dtype = int), dtype = int)
            mul_ind = np.vstack([mul_ind, np.eye(dim, dtype = int)])
            I = np.eye(dim, dtype = int)
            ind = [1] * dim
            for j in range(1,order):
                ind_new = []
                for i in range(dim):
                    a0 = np.copy(I[int(np.sum(ind[:i])):,:])
                    a0[:,i] += 1
                    mul_ind = np.vstack([mul_ind, a0])
                    ind_new += [a0.shape[0]]
                ind = ind_new
                I = np.copy(mul_ind[np.sum(q_num[:j+1]):,:])
            return mul_ind

        if trunc == 'TD':
            return TD(dim, order)
        elif trunc == 'TP':
            x = np.arange(order+1)
            return np.array(list(itls.product(x, repeat = dim)))[:,::-1]
        elif trunc == 'LQ':
            assert q<1 and q is not None, 'q must be less than 1 for L_q truncation.'
            mi = TD(dim, order)
            locs = np.where(np.sum(mi**q, axis = 1) <= order**q)
            return mi[list(locs),:][0,:,:]
    
    def mi_terms_loc(self, d1, d2, ord):
        assert d1 < d2
        MI = self.mi_terms(d2, ord)
        if d2 == d1 + 1:
            return np.where(MI[:,-1] == 0)[0]
        else:
            inds = np.vstack([np.where(MI[:,d1:] == [0]*(d2-d1))[0], np.where(MI[:,d1:] == [0]*(d2-d1))[1]]).T
            locs = []
            for i in range(MI.shape[0]):
                if len(np.where( inds[:,0] == i)[0]) == d2-d1:
                    locs += [i]
            return locs

